function! Hare_withSelection(existsCmd,emptyCmd)
  let visible = bufwinnr('^hare-selection$')
  if bufexists('hare-selection')
    set switchbuf+=useopen
    sbuf ^hare-selection$

    exe a:existsCmd

    if visible==-1 
      hide 
    else 
      sbuf % 
    endif
  else

    exe a:emptyCmd
    
  endif
endfunction

function! Hare_createSelection(entry)
  hide new hare-selection
  set bufhidden=hide
  set buftype=nofile
  set noswapfile
  map <buffer> <cr> :call Hare_sel_goto(0) <cr>
  call setline(1,a:entry)
  hide
endfunction

function! Hare_sel_clear()
  call Hare_withSelection(":bunload! hare-selection",":call confirm('empty selection')")
endfunction

function! Hare_sel_add_aux(entry)
  if nextnonblank(1)
    call append(line('$'),a:entry)
  else
    call setline(1,a:entry)
  endif
endfunction

function! Hare_sel_add()
  let l=line(".")
  let c=col(".")
  let entry="\"" . bufname("") . "\" " . l . " " . c . " : " . getline(l)
  let entry='"'.escape(entry,'"').'"'
  call Hare_withSelection(":call Hare_sel_add_aux(".entry.")",":call Hare_createSelection(".entry.")")
endfunction

function! Hare_sel_delete()
  let l=line(".")
  let c=col(".")
  let prefix="\"" . bufname("") . "\" " . l . " " . c . " : "
  call Hare_withSelection(":g/^".prefix."/d",":call confirm('empty selection')")
endfunction

function! Hare_sel_display()
  if bufexists('hare-selection')
    set switchbuf+=useopen
    sbuf ^hare-selection$
    normal J
    resize 5
  else
    call confirm('empty selection')
  endif
endfunction

function! Hare_sel_goto(motion)
  let visible = bufwinnr('^hare-selection$')
  if bufexists('hare-selection')
    set switchbuf+=useopen
    sbuf ^hare-selection$
    call cursor(line(".")+a:motion,0)
    let l=getline(".")
    let buffer=matchstr(l,"\"[^\"]*\" ")
    let buffer=strpart(buffer,1,strlen(buffer)-3)
    let i=matchend(l,"\"[^\"]*\" ")
    let l=strpart(l,i)
    let lnr=matchstr(l,"[0-9][0-9]*")
    let j=stridx(l," ")
    let cnr=matchstr(l,"[0-9][0-9]*",j)
    if visible==-1 
      hide 
    else 
      sbuf % 
    endif
    "call confirm(buffer ." ".lnr." ".cnr)
    exe ":sbuf " . buffer
    call cursor(lnr,cnr)
  else
    call confirm('empty selection')
  endif
endfunction

amenu Haskell(Refac).Selection.Clear\ selection :call Hare_sel_clear()<cr>
tmenu Haskell(Refac).Selection.Clear\ selection clear selection
amenu Haskell(Refac).Selection.Add\ to\ selection :call Hare_sel_add()<cr>
tmenu Haskell(Refac).Selection.Add\ to\ selection add current position to selection
amenu Haskell(Refac).Selection.Delete\ from\ selection :call Hare_sel_delete()<cr>
tmenu Haskell(Refac).Selection.Delete\ from\ selection delete current position from selection
amenu Haskell(Refac).Selection.Show\ selection :call Hare_sel_display()<cr>
tmenu Haskell(Refac).Selection.Show\ selection show selection
amenu Haskell(Refac).Selection.Visit\ next\ entry :call Hare_sel_goto(1)<cr>
tmenu Haskell(Refac).Selection.Visit\ next\ entry visit next entry
amenu Haskell(Refac).Selection.Visit\ current\ entry :call Hare_sel_goto(0)<cr>
tmenu Haskell(Refac).Selection.Visit\ current\ entry visit current entry
amenu Haskell(Refac).Selection.Visit\ previous\ entry :call Hare_sel_goto(-1)<cr>
tmenu Haskell(Refac).Selection.Visit\ previous\ entry visit previous entry

