

module RefacLocUtils(module HsTokens,PosToken, simpPos, 
                     SimpPos,unmodified,modified,simpPos0,ghead,glast,gfromJust,

                     tokenCol, tokenRow, tokenPos,tokenCon,tokenLen,lengthOfToks,
                     mkToken,defaultToken,newLnToken,whiteSpacesToken,whiteSpaceTokens,isWhite,
                     notWhite,isWhiteSpace,isNewLn,isCommentStart,isComment,
                     isNestedComment,isMultiLineComment,isOpenBracket,isCloseBracket,
                     isOpenSquareBracket,isCloseSquareBracket,isOpenBrace,isConid,
                     isLit,isWhereOrLet,isWhere,isLet,isIn,isCase,isDo,isIf,isForall,
                     isHiding,isModule,isComma,isEqual,isLambda,isIrrefute,isBar,isMinus,
                     endsWithNewLn,startsWithNewLn,hasNewLn,startsWithEmptyLn,setDefaultPos,
                     lastNonSpaceToken,firstNonSpaceToken,compressPreNewLns,compressEndNewLns,
                     lengthOfLastLine,getToks,updateExpToks,updateDeclToks,replaceToks',
                     replaceToks,deleteToks,adjustLayout,doRmWhites,doAddWhites,
                     mkTokensNoFstLnIndent,mkTokensWithIndent,srcLocs,-- getStartEndLoc,
                     startEndLoc,extendBothSides,extendForwards,extendBackwards,
                     startEndLocIncFowComment,startEndLocIncFowNewLn,startEndLocIncComments,
                     astToSource ,deleteFromToks, deleteFromToks1,
                     adjustOffset, -- * this function should be hided *
                     addParamInToks,  -- try to remove it 
                     StartEndLoc, updatePatToks,mkTokensNoFstLnIndent1,isArrow,swapInToks,
                     commentToks
                    
  ) where

import RefacTypeSyn(SimpPos)
import PosSyntax
import UniqueNames 
import HsLexerPass1 hiding (notWhite)
import HsTokens
import HsLayoutPre (PosToken)
import PrettyPrint
import HsExpUtil
import PNT 

import RefacTypeSyn
import Maybe
import List
import SourceNames
-------------------------
import DriftStructUtils  
import StrategyLib
------------------------

--In the token stream, locations are unique except the default locs.

{- Some related data types defined by Programatica's Lexer:
type Lexer = String -> LexerOutPut

type LexerOutput = [PosToken]

type PosToken = (Token,(Pos,String))

data Pos = Pos { char, line, column :: !Int } deriving (Show)
-- it seems that the field char is used to handle special characters including the '\t'

data Token
  = Varid | Conid | Varsym | Consym
  | Reservedid | Reservedop  | Specialid
  | IntLit | FloatLit | CharLit | StringLit
  | Qvarid | Qconid | Qvarsym | Qconsym
  | Special | Whitespace
  | NestedCommentStart -- will cause a call to an external function
  | NestedComment  -- from the external function
  | Commentstart  -- dashes
  | Comment -- what follows the dashes
  | ErrorToken | GotEOF | TheRest
  | ModuleName | ModuleAlias -- recognized in a later pass
  -- Inserted during layout processing (see Haskell 98, 9.3):
  | Layout     -- for implicit braces
  | Indent Int -- <n>, to preceed first token on each line
  | Open Int   -- {n}, after let, where, do or of, if not followed by a "{"
  deriving (Show,Eq,Ord)
-}

--A flag used to indicate whether the token stream has been modified or not.
unmodified = False
modified   = True

--- some default values----
pos0=Pos 0 0 0
simpPos0 = (0,0)

------------------------------------------------
ghead info (h:_) = h
ghead info []    = error $ "ghead "++info++" []"

glast info []      = error $ "glast " ++ info ++ " []"
glast info h@(_:_) = last h

gfromJust info (Just h) = h
gfromJust info Nothing = error $ "gfromJust " ++ info ++ " Nothing"

--Some functions for fetching a specific field of a token
tokenCol (_,(Pos _ _ c,_)) = c  

tokenRow (_,(Pos _ r _,_)) = r

tokenPos (_,(p,_))     = simpPos p

tokenCon (_,(_,s))     = s

tokenLen (_,(_,s))     = length s   --check this again! need to handle the tab key.

lengthOfToks::[PosToken]->Int
lengthOfToks=length.(concatMap tokenCon)

--Compose a new token using the given arguments.
mkToken::Token->SimpPos->String->PosToken
mkToken t (row,col) c=(t,(Pos 0 row col,c))

---Restriction: the refactorer should not modify refactorer-modified/created tokens.
defaultToken = (Whitespace, (pos0," "))      
newLnToken   = (Whitespace, (pos0,"\n"))  

--Create a list of white space tokens.
whiteSpacesToken::SimpPos->Int->[PosToken]
whiteSpacesToken (row,col) n
  |n>0        = [(Whitespace,(Pos 0 row col,replicate n ' '))]
  |otherwise  = []

--Some functions for checking whether a token is of a specific type of token.
isWhite (t,_)                = t==Whitespace || t==Commentstart || t==Comment || t==NestedComment
notWhite  = not.isWhite

isWhiteSpace (t,(_,s))       = t==Whitespace && s==" "
isNewLn (t,(_,s))            = t==Whitespace && s=="\n"

isCommentStart (t,(_,s))     = t==Commentstart && s=="--" 
isComment (t,(_,s))          = t==Comment || t ==NestedComment
isNestedComment (t,(_,s))    = t==NestedComment
isMultiLineComment (t,(_,s)) = t==NestedComment && (isJust (find (=='\n') s))

isOpenBracket  (t,(_,s))       = t==Special && s=="("
isCloseBracket (t,(_,s))       = t==Special && s==")"

isOpenSquareBracket  (t,(_,s)) = t==Special && s=="[" 
isCloseSquareBracket (t,(_,s)) = t==Special && s=="]"

isOpenBrace  (t,(_,s))         = t==Special && s=="{" 
isCloseBrace (t,(_,s))         = t==Special && s=="}" 

isConid (t,(_,_))              = t==Conid
isLit (t,(_,s)) = t==IntLit || t==FloatLit || t==CharLit || t==StringLit

isWhereOrLet  t   = isWhere t || isLet t
isWhere (t,(_,s)) = t==Reservedid && s=="where"
isLet   (t,(_,s)) = t==Reservedid && s=="let"

isIn    (t,(_,s))  = t==Reservedid && s=="in" 
isCase  (t,(_,s))  = t==Reservedid && s=="case"  
isDo    (t,(_,s))  = t==Reservedid && s=="do"  
isIf    (t,(_,s))  = t==Reservedid && s=="if"
isForall (t,(_,s)) = t==Reservedid && s=="forall"
isHiding (t,(_,s)) = s=="hiding" 
isModule (t,(_,s)) = t==Reservedid && s=="module"
  
isComma (t,(_,s))    = t==Special && s=="," 
isEqual  (t,(_,s))   = t==Reservedop && s=="="
isLambda (t,(_,s))   = t==Reservedop && s=="\\"
isIrrefute (t,(_,s)) = t==Reservedop && s=="~"
isBar   (t,(_,s))    = t==Reservedop && s=="|"
isArrow (t,(_,s))    = t==Reservedop && s=="->"
isMinus (t,(_,s))    = t==Varsym && s=="-"
-----------------------------------------------------------------

--Returns True if the token ends with '\n'
endsWithNewLn::PosToken->Bool
endsWithNewLn   (_,(_,s)) =if s==[] then False 
                                    else (glast "endsWithNewLn"  s=='\n')
 
--Returns True if the token starts with `\n`.
startsWithNewLn::PosToken->Bool
startsWithNewLn (_,(_,s)) =if s==[] then False 
                                    else ((ghead "starsWithNewLn" s)=='\n')

--Returns True if there is a '\n' in the token.
hasNewLn::PosToken->Bool
hasNewLn (_,(_,s))=isJust (find (=='\n') s) {-compare with: isNewLn -}


--Returns True if a token stream starts with a newline token (apart from the white spaces tokens)
startsWithEmptyLn::[PosToken]->Bool
startsWithEmptyLn toks=isJust (find isNewLn $ takeWhile (\t->isWhiteSpace t || isNewLn t) toks)

-- CHECK WHETHER THIS FUNCTION CAN BE REMOVED.
--Set the all the positions in a token stream to pos0
setDefaultPos::[PosToken]->[PosToken]
setDefaultPos toks =map (\(t,(_,s))->(t,(pos0,s))) toks 

-- get the last non-space token in a token stream.
lastNonSpaceToken::[PosToken]->PosToken
lastNonSpaceToken toks=case dropWhile isWhiteSpace (reverse toks) of
                         [] ->defaultToken
                         l -> ghead "lastNonSpaceToken" l

--get the first non-space token in a token stream.
firstNonSpaceToken::[PosToken]->PosToken
firstNonSpaceToken toks=case dropWhile isWhiteSpace toks of
                         [] ->defaultToken
                         l -> ghead "firstNonSpaceToken" l  

-- remove the preceding extra empty lines.
compressPreNewLns::[PosToken]->[PosToken]
compressPreNewLns toks= let (toks1, toks2) = break (not.(\t->isNewLn t || isWhiteSpace t)) toks
                            groupedToks    = groupBy (\x y->tokenRow x==tokenRow y) toks1
                        in if length groupedToks>1 then (last groupedToks)++toks2  
                                                   else toks

--remove the following extra empty lines.
compressEndNewLns::[PosToken]->[PosToken]
compressEndNewLns toks=let (toks1, toks2) = break (not.(\t->isNewLn t || isWhiteSpace t)) (reverse toks)
                           groupedToks    = groupBy (\x y->tokenRow x==tokenRow y) toks1
                       in if length groupedToks>1 then reverse ((ghead "compressEndNewLns" groupedToks)++toks2) 
                                                  else toks


--Give a token stream covering multi-lines, calculate the length of the last line  
lengthOfLastLine::[PosToken]->Int
lengthOfLastLine toks=let (toks1,toks2)=break hasNewLn $ reverse toks
                      in  if toks2==[] then sum (map tokenLen toks1)
                                       else sum (map tokenLen toks1)+lastLineLenOfToken (ghead "lengthOfLastLine"  toks2)
  where 
   --Compute the length of a token, if the token covers multi-line, only count the last line.
   --What about tab keys?
   lastLineLenOfToken (_,(_,s))=(length.(takeWhile (\x->x/='\n')).reverse) s


--get a token stream specified by the start and end position.
getToks::(SimpPos,SimpPos)->[PosToken]->[PosToken]
getToks (startPos,endPos) toks
    =let (_,toks2)=break (\t->tokenPos t==startPos) toks
     in fst $ break (\t->tokenPos t>endPos) toks2 


--comment a token stream specified by the start and end position.
commentToks::(SimpPos,SimpPos)->[PosToken]->[PosToken]
commentToks (startPos,endPos) toks
    = let (toks1,toks2)=break (\t->tokenPos t==startPos) toks
          (toks21,toks22)=break (\t->tokenPos t==endPos) toks2
          toks21' = case  toks21 of
                          [] -> []
                          ((t,(l,s)):ts) -> (t,(l, ("{- "++s))):ts
          toks22' = case toks22 of 
                          [] -> []
                          ((t,(l,s)):ts) -> (t, (l, (s++" -}"))):ts
      in toks1++toks21'++ toks22'
      
--replace oldExp (SHOULD BE AN EXPRESSION) by newExp in the token stream.

updateExpToks oldExp newExp 
   = do ((toks,_), others) <-get
        let (startPos, endPos) = getStartEndLoc toks oldExp
            (toks1,toks2)=break (\t->tokenPos t>=startPos) toks
            offset=lengthOfLastLine toks1
            newToks = mkTokensNoFstLnIndent1 Varid startPos offset  (astToSource newExp)
            toks' = replaceToks toks startPos endPos newToks
        put ((toks',modified), others)


--replace oldExp (SHOULD BE AN EXPRESSION) by newExp in the token stream.
updatePatToks oldPats newPats
   = do ((toks,_), others) <-get
        let (startPos, endPos) = getStartEndLoc toks oldPats
            (toks1,toks2)=break (\t->tokenPos t>=startPos) toks
            offset=lengthOfLastLine toks1
            newToks = mkTokensNoFstLnIndent1 Varid startPos offset $ [format newPats]
            toks' = replaceToks toks startPos endPos newToks
        put ((toks',modified), others)
  where 
   format [] = "" 
   format [p] = (render.ppi) p
   format (p:ps) = (render.ppi) p ++" " ++ format ps


--replace one declaration by another declaration
updateDeclToks oldDecl newDecl
  = do ((toks, _),others) <-get
       let (startPos, endPos) = getStartEndLoc toks oldDecl
           -- pretty-print the whole data declaration.
           newToks = mkTokensWithIndent Varid startPos (astToSource newDecl) True
           toks'= replaceToks toks startPos endPos newToks        
       put ((toks', modified),others)



swapInToks t1 t2
 = do ((toks,_),others)<-get
      let (startPos1,endPos1)= getStartEndLoc toks t1
          (startPos2,endPos2)= getStartEndLoc toks t2
          t1Toks=getToks (startPos1,endPos1) toks
          t2Toks=getToks (startPos2,endPos2) toks
          toks'=replaceToks toks startPos2 endPos2 t1Toks
          toks''=replaceToks toks' startPos1 endPos1 t2Toks
      put ((toks'', modified), others)

addParamInToks t newParams formalParam 
  = do ((toks,_),others)<-get   
       let (startPos,endPos) = getStartEndLoc toks t
           (_,toks2) = break (\t->tokenPos t == startPos) toks
           tToks     = takeWhile (\t->tokenPos t<=endPos) toks2
           newToken  =if formalParam  
                      then mkToken Varid startPos $ foldl (\x y->x++" "++y) (concatMap tokenCon tToks) newParams
                      else mkToken Varid startPos $ "("++(foldl (\x y->x++" "++y) (concatMap tokenCon tToks) newParams) ++")"
           toks' = replaceToks toks startPos endPos [newToken]
       put ((toks',modified),others)
                  
--Replace the list of tokens which starts at `startPos` and ends at `endPos` by the new token list `newToks`
replaceToks'::((MonadState (([PosToken], Bool), t) m),Monad m)=>SimpPos->SimpPos->[PosToken]->m ()
replaceToks' startPos endPos newToks
  = do ((toks,_),others)<-get                           
       let toks'=replaceToks toks startPos endPos newToks
       put ((toks',modified),others)

--Replace a list of tokens in the token stream by a new list of tokens, adjust the layout as well.
--To use this function make sure the start and end positions really exist in the token stream.
--QN: what happens if the start or end position does not exist?

replaceToks::[PosToken]->SimpPos->SimpPos->[PosToken]->[PosToken]

replaceToks toks startPos endPos newToks
   = let (toks1,toks2)=break (\t->tokenPos t>=startPos ) toks
          --toks21: the tokens to be replaced.
          --toks1: the tokens before toks21
          --toks22: the tokens after toks21
         (toks21,toks22)=break (\t->tokenPos t>endPos) toks2
         -- CHECK THIS: IS THE nextPos correct for adjust layout?
         nextPos=if toks22==[] then endPos else tokenPos(ghead "replaceToks"  toks22)  -- the position that is right after endPos.
     in  case (isSingleLine toks21, isSingleLine newToks) of
                 (True,True)   -> replaceToksCase1 toks1 toks21 toks22 nextPos  
                 (True,False)  -> replaceToksCase2 toks1 toks21 toks22 nextPos  
                 (False,True)  -> replaceToksCase3 toks1 toks21 toks22 nextPos  
                 (False,False) -> replaceToksCase4 toks1 toks21 toks22 nextPos 
     where 
         
         --'singleLine' test whether all tokens in 'toks' are in the same line.
         isSingleLine toks=isNothing $ find hasNewLn toks

          --replace toks in a single line by toks in a single line. 
         replaceToksCase1 toks1 toks21 toks22 nextPos
            =let oldToksLength=lengthOfToks toks21   
                 newToksLength=lengthOfToks newToks 
             in  toks1++newToks++adjustLayout (newToksLength-oldToksLength) nextPos toks22 

         --replace toks in a single line by toks in multi lines.
         replaceToksCase2 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks1+lengthOfToks toks21
                newOffset=lengthOfLastLine newToks 
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22

         --replace toks in multi lines by toks in a single line. 
         replaceToksCase3 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks21 
                newOffset=lengthOfLastLine toks1+lengthOfToks newToks 
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22

         --replace toks in multi lines by toks in multi lines.  
         replaceToksCase4 toks1 toks21 toks22 nextPos
           =let oldOffset=lengthOfLastLine toks21 
                newOffset=lengthOfLastLine newToks                   
            in toks1++newToks++adjustLayout (newOffset-oldOffset) nextPos toks22

{- Delete an syntax phrase from the token stream, this function (instead of the following one)
   should be the interface function for deleting tokens.
-}          

-- deleteFromToks::( (MonadState (([PosToken], Bool), t1) m), StartEndLoc t,Printable t,Term t)=>t->m ()
deleteFromToks t 
   =do ((toks,_),others)<-get 
       let (startPos,endPos)=getStartEndLoc toks t
           toks'=deleteToks toks startPos endPos
       put ((toks',modified),others)

-- deleteFromToks::( (MonadState (([PosToken], Bool), t1) m), StartEndLoc t,Printable t,Term t)=>t->m ()
deleteFromToks1 t 
   =do ((toks,_),others)<-get 
       let (startPos,endPos)=startEndLocIncFowNewLn toks t
           toks'=deleteToks toks startPos endPos
       put ((toks',modified),others)
         
{-Delete a sequence of tokens specified by the start position and end position from the token stream,
  then adjust the remaining token stream to preserve layout-}
deleteToks::[PosToken]->SimpPos->SimpPos->[PosToken]
deleteToks toks startPos@(startRow, startCol) endPos@(endRow, endCol)
    =let (toks1,toks2)=break (\t->tokenPos t>=(startRow,0)) toks
         (toks21,toks22)=break (\t->tokenPos t>=(endRow+1, 0)) toks2
         --'before'is the tokens preceeding the tokens to be deleted at 'startRow'
         before=takeWhile (\t->tokenPos t<startPos) toks21  
         --'after' is the tokens following the tokens to be deleted at 'endRow'
         after =dropWhile (\t->tokenPos t<=endPos) toks21
     in if after==[]   
          then  let toks22'=if isMultiLineComment (lastNonSpaceToken toks21) && --handle deleting multi-line comment
                              toks22/=[] && tokenCol (ghead "deleteToks1" toks22)>1  
                             then  let (row,col)=tokenPos $ ghead "deleteToks2" toks22   
                                   in whiteSpacesToken (row,1) (col-1) ++ toks22 
                             else toks22
                    toks1AndBefore=toks1++before
                in if (toks1AndBefore/=[] && endsWithNewLn (last toks1AndBefore)) 
                      || (toks22'/=[] && startsWithNewLn (ghead "deleteToks3" toks22'))
                      then  toks1++before++toks22'
                      --avoiding layout adjustment by adding a `\n', sometimes may produce extra lines.
                      else  toks1++before++[newLnToken]++toks22'
          else let nextTok=ghead "deleteToks4" after  
               in toks1++before++adjustLayout (startCol-tokenCol nextTok) (tokenPos nextTok) (after++toks22) 

  
{-adjust the token stream to reflect the layout change because of adding/removing tokens
  Arguments: Int : the offest. SimPos : the position from where the following tokens will be afftected.
-}
adjustLayout::Int->SimpPos->[PosToken]->[PosToken]
adjustLayout 0 _ toks = toks   -- No adjustment is need.    
adjustLayout n pos toks 
   =if layoutRuleApplies pos toks 
      then if n>0 
             then {-start adjusting from next row by adding n whitespaces at the beginnig of each row affected-}
                  addSpaces n (snd pos) toks
             else {-start adjusting from next row by removing n whitespaces from the beginning of each row affected-} 
                  rmSpaces (-n)(snd pos) toks  
      else toks --no need to adjust layout.

    where 

      {-Test whether there is a layout-rule-invoking token (i.e.'where','let','do' or 'of') in the current
        line after position 'pos'. -}
      layoutRuleApplies::SimpPos->[PosToken]->Bool
      layoutRuleApplies pos toks
         = --t is the remained toks in current line from position 'pos'
           let t=dropWhile (not.isSpecialToken) $ filter notWhite $ takeWhile (not.hasNewLn) toks
           in length t>1 && tokenCon (t!!1)/="{"           
         where 
            isSpecialToken t=elem (tokenCon t) ["where","let","do","of"]            
  
      {-add/remove n whitespace tokens to/from each row affected by the current layout in the remaining token list.  
       n: the number of spaces to be added/removed
       col is used to determine whether the adjusting should continue.-}
      addSpaces,rmSpaces::Int->Int->[PosToken]->[PosToken]

      addSpaces n col toks
          =let groupedToks=groupBy (\x y->tokenRow x==tokenRow y) toks
           in  if groupedToks==[] 
                  then [] 
                  else ghead "addSpaces" groupedToks ++ concatMap (insertWhites n col) (tail groupedToks) 
  
      rmSpaces n col toks
         =let groupedToks=groupBy (\x y->tokenRow x==tokenRow y) toks
          in if groupedToks==[]
               then [] 
               else  ghead "rmSpaces" groupedToks ++ concatMap (rmWhites n col) (tail groupedToks)           
      
      --Remove n whitespaces from the beginning.
      --Assumption: assume there are at least n whitespaces.
      --PROBLEM: This Assumption can not always be guaranteed if
      --comments are allowed to be followed by code in the same line.
      rmWhites n col toks
        = let t=firstNotWhiteTok toks
          in if t/=defaultToken && tokenCol t>col
              then doRmWhites n toks
              else toks 
       --Insert n whitespace tokens at the beginning. 
      insertWhites n col toks
        = let t=firstNotWhiteTok toks
          in if t/=defaultToken && tokenCol t>=col
                then doAddWhites n toks
                else toks 

      --Get the first not-white token.
      firstNotWhiteTok toks 
          = fromMaybe defaultToken (find notWhite toks)
  
-- remove at most n white space tokens from the beginning of ts
doRmWhites::Int->[PosToken]->[PosToken]
doRmWhites 0 ts=ts  
doRmWhites n []=[]
doRmWhites n toks@(t:ts)=if isWhiteSpace t then doRmWhites (n-1) ts
                                           else toks 

--add n white space tokens to the beginning of ts
doAddWhites::Int->[PosToken]->[PosToken]
doAddWhites n []=[]
doAddWhites n ts@(t:_)= whiteSpacesToken (tokenRow t,0) n ++ts


--Starting from the second line, add/remove 'abs offset' number of white spaces before each line.
--if offset>0 then add white spaces, otherwise remove white spaces.

adjustOffset::Int->[PosToken]->Bool->[PosToken]           
adjustOffset offset [] _ = []  
adjustOffset offset toks firstLineIncluded  
     = let groupedToks = groupBy (\x y->tokenRow x==tokenRow y) toks  --groupedToks/=[], no problem with 'head'
           --if firstLineIncluded is False, the offset of the first line won't be ajusted.
       in if offset>=0 then if firstLineIncluded 
                               then concatMap (doAddWhites offset) groupedToks
                               else ghead "adjustOffset" groupedToks ++ concatMap (doAddWhites offset) (tail groupedToks)
                       else if firstLineIncluded 
                               then concatMap (doRmWhites  (-offset)) groupedToks
                               else ghead "adjustOffset" groupedToks ++ concatMap (doRmWhites  (-offset)) (tail groupedToks) 


--Given several lines of source, put each line into one token, and the start location is (row,col).
--Except the first line, (col-1) number of white spaces will be added to the beginning of each line.
mkTokensNoFstLnIndent::Token->SimpPos->[String]->[PosToken]
mkTokensNoFstLnIndent  tokenType (row,col) []=[]
mkTokensNoFstLnIndent  tokenType (row,col) (line:lines)
    =if lines==[] then [mkToken tokenType (row,col) line]
                  else [(mkToken tokenType (row,col) line),newLnToken]++
                       mkTokensWithIndent tokenType (row,col) lines False


mkTokensNoFstLnIndent1::Token->SimpPos->Int->[String]->[PosToken]
mkTokensNoFstLnIndent1  tokenType (row,col) offset []=[]
mkTokensNoFstLnIndent1  tokenType (row,col) offset (line:lines)
    =if lines==[] then [mkToken tokenType (row,col) line]
                  else [mkToken tokenType (row,col) (line++"\n"++ mkSourceWithIndent tokenType offset lines False)]
--Given several lines of source, put each line into one token, and for each line, (col-1)
--number of white spaces will be added to the beginning of each line. if addSpace is True,
-- a white space will be added to the end of the first line.
mkSourceWithIndent::Token->Int->[String]->Bool->String
mkSourceWithIndent tokenType offset [] addSpace=""
mkSourceWithIndent tokenType offset (line:lines) addSpace
  = if lines/=[]  
       then  replicate offset ' '++line++"\n"
              ++ mkSourceWithIndent tokenType offset lines addSpace
       else  if addSpace
                then  replicate offset ' '++line++"  "
                else  replicate offset ' '++line
  

--Given several lines of source, put each line into one token, and for each line, (col-1)
--number of white spaces will be added to the beginning of each line. if addSpace is True,
-- a white space will be added to the end of the first line.
mkTokensWithIndent::Token->SimpPos->[String]->Bool->[PosToken]
mkTokensWithIndent tokenType (row,col) [] addSpace=[]
mkTokensWithIndent tokenType (row,col) (line:lines) addSpace
  = if lines/=[]  
       then  lineToTokens tokenType (row,col) line
              ++mkTokensWithIndent tokenType (row,col) lines addSpace
       else  if addSpace
                then  whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) (line++"  ")]
                else  whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) line]
   where 
        lineToTokens tokenType (row,col) line
           =whiteSpaceTokens (row,1) (col-1) ++ [mkToken tokenType (row,col) line,
                 mkToken Whitespace (row, col+length line) "\n"]  

whiteSpaceTokens (row, col) n 
 = if n<=0 then [] 
           else (mkToken Whitespace (row,col) " "):whiteSpaceTokens (row,col+1) (n-1)

           
-------------------------------------------------------------------------------------------------
--get all the source locations (use locations) in an AST phrase t, and sort it in an ascending order.
srcLocs::(Term t)=> t->[SimpPos]
srcLocs t =(sort.nub.srcLocs') t \\ [simpPos0]
   where srcLocs'=runIdentity.(applyTU (full_tdTU (constTU []
                                                  `adhocTU` pnt  
                                                  `adhocTU` literalInExp
                                                  `adhocTU` literalInPat)))
        
         pnt (PNT pname _ (N (Just (SrcLoc _  _ row col))))=return [(row,col)]                
         pnt _=return []
                     
         literalInExp ((Exp (HsLit (SrcLoc _  _ row col) _))::HsExpP) = return [(row,col)]
         literalInExp (Exp _) =return []

         literalInPat ((Pat (HsPLit (SrcLoc _ _ row col) _))::HsPatP) = return [(row,col)]
         literalInPat (Pat (HsPNeg (SrcLoc _  _ row col) _)) = return [(row,col)]
         literalInPat _ =return []  

--------------------------------------------------------------------------------------------------------
-- This function should be the interface function for fetching start and end locations of a AST phrase in the source.


getStartEndLoc::(Term t, StartEndLoc t,Printable t)=>[PosToken]->t->(SimpPos,SimpPos)
getStartEndLoc toks t 
    =let locs = srcLocs  t
         (startPos',endPos') = startEndLoc toks t
         (startPos,endPos) =if locs == [] 
                               then (startPos', endPos') -- still can not handle [], () and _
                               else  -- tries to correctify the problems cause by [],() and _
                                   (if startPos' == simpPos0 then ghead "getStartEndLoc" locs else startPos', 
                                    if endPos'   == simpPos0 then last locs else endPos')
     in if startPos'/=simpPos0 && endPos'==simpPos0  -- this expression ends (but not begins) with [], () or _,
          then adjustLoc toks (startPos,endPos) t
          else (startPos,endPos)
    where  
      adjustLoc toks (startPos,endPos) t
         = let subToks = getToks (startPos,endPos) toks  -- get the tokens from the token stream.
               tokenStr = concatMap tokenCon $ filter (\t->not (isWhite t || isCloseBracket t || isOpenBracket t)) subToks
               astStr   = filter (\c->c /=' ' && c/='(' && c/=')' && c/='\n' && c/='\t') $ (render.ppi) t
           in  if length astStr > length tokenStr  
                 then let toks1 = dropWhile (\t->tokenPos t <= endPos) toks
                      in if toks1/=[] 
                            then adjustLoc toks (startPos, tokenPos (ghead "adjustLoc" toks1)) t 
                            else (startPos,endPos)                     
                 else (startPos,endPos)
            
{-
getStartEndLoc::(Term t, StartEndLoc t,printable t)=>[PosToken]->t->(SimpPos,SimpPos)
getStartEndLoc toks t
    = let locs = srcLocs  t
          (startPos,endPos) = startEndLoc toks t
      in if locs == [] 
           then (startPos, endPos) -- still can not handle [], () and _
           else  -- tries to correctify the problems cause by [],() and _
               (if startPos == simpPos0 then head locs else startPos, 
                if endPos   == simpPos0 then last locs else endPos)           

-}
--given an AST phrase, 'startEndLoc' gets its start and end position in the program source.
class StartEndLoc t where

   startEndLoc :: [PosToken]->t->(SimpPos,SimpPos)


instance StartEndLoc HsModuleP where

  startEndLoc toks _  = (tokenPos (ghead "startEndLoc:HsModuleP" toks), 
                         tokenPos (glast "startEndLoc:HsModuleP" toks))
instance StartEndLoc HsExpP where
    
  startEndLoc  toks (Exp e)=
      case e of
    
        HsId ident@(HsVar (PNT pn _ _)) -> let (startLoc, endLoc) = startEndLoc toks ident
                                               {- To handle infix operator. for infix operators like (++), there
                                                  is no parenthesis in the syntax tree -}
                                               (toks1,toks2) = break (\t->tokenPos t==startLoc) toks
                                               toks1' = dropWhile isWhite (reverse toks1)
                                               toks2' = dropWhile isWhite (dropWhile (\t->tokenPos t <=endLoc) toks2)
                                           in if toks1'/=[] && toks2'/=[] && isOpenBracket (head toks1')
                                                 && isCloseBracket (head toks2')
                                              then (tokenPos (head toks1'), tokenPos (head toks2'))
                                              else (startLoc, endLoc) 
        HsId  x                       ->startEndLoc toks x
     

        HsLit (SrcLoc _ _ r c) _        -> ((r,c),(r,c))

        HsInfixApp e1 op e2           ->let (startLoc,_)=startEndLoc toks e1
                                            (_, endLoc) =startEndLoc toks e2
                                        in (startLoc,endLoc)

        e@(HsApp e1 e2)               ->let (startLoc,endLoc)=startEndLoc toks e1
                                            (startLoc1, endLoc1 )=startEndLoc toks e2
                                        in (startLoc, endLoc1)

        HsNegApp (SrcLoc _ _ r c) e     ->let (_,endLoc)=startEndLoc toks e
                                          in ((r,c), endLoc)

        HsLambda ps e                 ->let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsLambda" ps)  --ps can not be empty
                                            (_,endLoc)  =startEndLoc toks e
                                        in extendForwards toks startLoc endLoc isLambda

        HsIf e1 e2 e3                 ->let (startLoc, _)=startEndLoc toks e1
                                            (_, endLoc)=startEndLoc toks e3
                                        in extendForwards toks  startLoc endLoc isIf

        HsLet ds e                    ->if ds==[]
                                          then
                                            let  (startLoc,endLoc)=startEndLoc toks e
                                            in extendForwards toks startLoc endLoc isLet
                                          else
                                            let  (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsLet" ds) 
                                                 (_,endLoc)  =startEndLoc toks e 
                                            in extendForwards toks startLoc endLoc isLet

        HsCase e alts                 ->let (startLoc,_)=startEndLoc toks e
                                            (_,endLoc)  =startEndLoc toks (glast "HsCase" alts) --alts can not be empty.
                                        in extendForwards toks startLoc endLoc isCase
                        
        HsDo stmts                    ->let (startLoc, endLoc)=startEndLoc toks  stmts
                                        in extendForwards toks startLoc endLoc isDo

        HsTuple es                    ->if es==[] 
                                         then (simpPos0,simpPos0)  --Empty tuple can cause problem.
                                         else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsTuple" es)
                                                  (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsTuple" es)
                                              in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  

        HsList es                     ->if es==[] 
                                         then (simpPos0,simpPos0)  --Empty list can cause problem.
                                         else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsList" es) 
                                                  (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsList" es)
                                              in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket

        HsParen e                     ->let (startLoc,endLoc)=startEndLoc toks  e
                                        in if expIsPNT e
                                             then (startLoc, endLoc)
                                             else extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket 
                                          where
                                            expIsPNT (Exp (HsId (HsVar pnt)))=True
                                            expIsPNT (Exp (HsParen e))=expIsPNT e  
                                            expIsPNT _ =False

        HsLeftSection e op            ->let (startLoc,_)=startEndLoc toks e
                                            (_, endLoc )=startEndLoc toks op
                                        in (startLoc,endLoc)

        HsRightSection op e           ->let (startLoc,_)=startEndLoc toks op
                                            (_, endLoc )=startEndLoc toks op
                                        in (startLoc,endLoc)

        HsRecConstr i upds            ->let (startLoc,_)=startEndLoc toks i
                                            (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsRecConstr" upds) --can 'upds' be empty?
                                        in extendBackwards toks startLoc endLoc isCloseBrace 
                                        
        HsRecUpdate e upds            ->let (startLoc,_)=startEndLoc toks e
                                            (_,endLoc)  =startEndLoc toks (glast "startEndLoc:HsRecUpdate" upds) --ditto
                                        in extendBackwards toks startLoc endLoc isCloseBrace 

        HsEnumFrom e                  ->let (startLoc,endLoc)=startEndLoc toks e
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket   

        HsEnumFromTo e1 e2            ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e2
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket   

        HsEnumFromThen e1 e2          ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e2
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket  
                                  
        HsEnumFromThenTo e1 e2 e3     ->let (startLoc,_)=startEndLoc toks e1
                                            (_,  endLoc)=startEndLoc toks e3
                                        in extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket  
                                 
        HsListComp stmts              ->let (startLoc,endLoc)=startEndLoc toks stmts
                                        in  extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket     

        HsAsPat i e                   ->let (startLoc,_)=startEndLoc toks i
                                            (_,endLoc)=  startEndLoc toks e
                                        in (startLoc,endLoc) 

        HsIrrPat e                    ->let (startLoc,endLoc)=startEndLoc toks e
                                        in extendForwards toks startLoc endLoc isIrrefute

        HsWildCard                    ->(simpPos0,simpPos0)  -- wildcard can cause problem.
  
        
        HsExpTypeSig loc e c t        ->let (startLoc,_)=startEndLoc toks e
                                            (_, endLoc )=startEndLoc toks t
                                        in (startLoc,endLoc)    

instance StartEndLoc HsTypeP where

   startEndLoc toks (Typ p)=
      case p of 
        HsTyFun  t1  t2       ->   let (startLoc,_)=startEndLoc toks t1
                                       (_ , endLoc)=startEndLoc toks t2
                                   in (startLoc,endLoc)  
        --HsTyTuple [t]         ->

        HsTyApp  t1 t2        ->   let (startLoc,_)=startEndLoc toks  t1 
                                       (_ , endLoc)=startEndLoc toks  t2
                                   in (startLoc,endLoc)

        HsTyVar  i            ->   startEndLoc toks i
        HsTyCon  i            ->   startEndLoc toks i

        HsTyForall   is  t   ->   case is of
                                     []  ->let (startLoc,endLoc)=startEndLoc toks t
                                           in  extendForwards toks startLoc endLoc isForall
   
                                     l  -> let (startLoc, _) =startEndLoc toks  $ ghead "StartEndLoc:HsTypeP" is
                                               ( _ , endLoc) =startEndLoc toks t
                                           in extendForwards toks startLoc endLoc isForall
instance StartEndLoc HsPatP where

  startEndLoc toks (Pat p)=
     case p of
       HsPId i                          ->startEndLoc toks i

       HsPLit (SrcLoc _ _ r c) _        ->((r,c),(r,c))

       HsPNeg (SrcLoc _ _  r c) p       ->((r,c),(r,c))
   
       HsPInfixApp p1 op p2             ->let (startLoc,_)=startEndLoc toks  p1
                                              (_ , endLoc)=startEndLoc toks p2
                                          in (startLoc,endLoc)

       HsPApp i ps                      ->let (startLoc,_)=startEndLoc toks  i
                                              (_,endLoc)=startEndLoc toks (glast "StartEndLoc:HsPatP" ps)
                                          in (startLoc,endLoc)

       HsPTuple ps                      -> if ps==[]
                                             then  (simpPos0,simpPos0)  --Check this
                                             else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsPTuple"  ps)
                                                      (_,endLoc)=startEndLoc toks (glast "startEndLoc:HsPTuple" ps)
                                                  in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  

       HsPList ps                       ->if ps==[] 
                                            then (simpPos0,simpPos0)  --Check this
                                            else let (startLoc,_)=startEndLoc toks (ghead "startEndLoc:HsPList" ps)
                                                     (_, endLoc) =startEndLoc toks (glast "startEndLoc:HsPList" ps)
                                            in  extendBothSides toks startLoc endLoc isOpenSquareBracket isCloseSquareBracket     
                                        
       HsPParen p                       ->let (startLoc,endLoc)=startEndLoc toks p
                                          in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket  
 
       HsPRec i upds                    ->let (startLoc,_)=startEndLoc toks i
                                              (_,endLoc)=startEndLoc toks (glast "startEndLoc:HsPRec" upds) --can upds be empty?
                                          in extendBackwards toks startLoc endLoc isCloseBrace 
                                         
       HsPAsPat i p                     ->let (startLoc,_)=startEndLoc toks i
                                              (_,endLoc)=startEndLoc toks p
                                          in (startLoc,endLoc)
 
       HsPIrrPat p                      ->let (startLoc,endLoc)=startEndLoc toks p
                                              toks1=takeWhile (\t->tokenPos t<startLoc) toks  
                                          in extendForwards toks startLoc endLoc isIrrefute
                                       
       HsPWildCard                       ->(simpPos0,simpPos0)  -- wildcard can  cause problem.

instance StartEndLoc [HsPatP] where

   startEndLoc toks ps = let locs=(nub.(map (startEndLoc toks))) ps \\ [(simpPos0,simpPos0)]
                         in if locs==[] then (simpPos0,simpPos0)
                                        else let (startLoc,_)=ghead "StartEndLoc:HsPatP" locs
                                                 (_,endLoc) =glast "StartEndLoc:HsPatP"  locs
                                             in (startLoc,endLoc)
instance StartEndLoc [HsExpP] where

   startEndLoc toks es=let locs=(nub.(map (startEndLoc toks))) es \\ [(simpPos0,simpPos0)]
                       in if locs==[] then (simpPos0,simpPos0)
                                      else let (startLoc,_)=ghead "StartEndLoc:HsExp" locs
                                               (_,endLoc) =glast "startEndLoc:HsExp" locs
                                           in (startLoc,endLoc)
instance StartEndLoc [HsDeclP] where
  startEndLoc toks ds=if  ds==[] then (simpPos0,simpPos0)
                                 else if length ds==1 
                                        then startEndLoc toks (ghead "StartEndLoc:[HsDeclP]" ds)
                                        else  let (startLoc,_)=startEndLoc toks (ghead "StartEndLoc:[HsDeclP]" ds)
                                                  (_,endLoc) =startEndLoc toks (glast  "StartEndLoc:[HsDeclP]" ds)
                                              in (startLoc,endLoc)

instance StartEndLoc HsMatchP where
   startEndLoc toks (HsMatch loc i ps rhs ds)
         =let (startLoc,_)=startEndLoc toks i
              (_,endLoc)  =if ds==[] then startEndLoc toks rhs
                                     else startEndLoc toks (glast "StartEndLoc:HsMatchP" ds)
          in (startLoc,endLoc)
                                  


instance StartEndLoc HsStmtP where      -- Bug fixed. 20/05/2004
   startEndLoc toks stmts=let s=getStmtList  stmts                           
                              locs = map (startEndLoc toks) s 
                              (startLocs, endLocs) =(sort (map fst locs), sort (map snd locs))
                          in (ghead "StartEndLoc::HsStmtP" startLocs, glast "StartEndLoc::HsStmtP" endLocs)
                              
instance StartEndLoc (HsStmtAtom HsExpP HsPatP [HsDeclP])  where

    startEndLoc toks stmt=
      case stmt of 
           HsGeneratorAtom (SrcLoc _ _ r c) p e -> 
                                  let (startLoc,_)=startEndLoc toks p
                                      (_,endLoc)  =startEndLoc toks e
                                  in (startLoc,endLoc)
           HsQualifierAtom e   -> startEndLoc toks e
           HsLetStmtAtom ds    -> if ds==[]
                                   then (simpPos0,simpPos0)
                                   else let (startLoc,_)= startEndLoc toks (ghead "StartEndLoc:HsStmtAtom" ds)
                                            (_,endLoc)  = startEndLoc toks (glast "StartEndLoc:HsStmtAtom" ds)
                                        in (startLoc,endLoc)
           HsLastAtom e        ->startEndLoc toks e

instance (StartEndLoc i,StartEndLoc e)=>StartEndLoc (HsFieldI i e) where
    startEndLoc toks (HsField i e)=let (startLoc,_)=startEndLoc toks i
                                       (_,endLoc)=startEndLoc toks e
                                   in (startLoc,endLoc) 

instance StartEndLoc HsAltP where   
    startEndLoc toks (HsAlt l p rhs ds)=let (startLoc,_)=startEndLoc toks p
                                            (_,endLoc)=if ds==[] then startEndLoc toks rhs
                                                                 else startEndLoc toks (glast "StartEndLoc:HsAltP" ds)
                                        in (startLoc,endLoc)

instance StartEndLoc RhsP where    
   startEndLoc toks (HsBody e)=startEndLoc toks e

   startEndLoc toks (HsGuard es)=if es==[] then (simpPos0,simpPos0)
                                           else let (_,e1,_)=ghead "StartEndLoc:RhsP" es
                                                    (_,_,e2)=glast "StartEndLoc:RhsP" es
                                                    (startLoc,_)=startEndLoc toks e1
                                                    (_,endLoc)=startEndLoc toks e2
                                                in (startLoc,endLoc) 

instance StartEndLoc (HsIdentI PNT) where
    startEndLoc toks ident =
       case ident of
           HsVar i  ->startEndLoc toks i
           HsCon i  ->startEndLoc toks i


instance StartEndLoc (HsImportDeclI ModuleName PNT)  where
     startEndLoc toks (HsImportDecl (SrcLoc _ _ row col) modName qual  as Nothing)
        = let startPos=(row,col)
              endPos = if isJust as then snd (startEndLoc toks (fromJust as))
                                    else snd (startEndLoc toks modName)
          in (startPos, endPos) 
                                        
     startEndLoc toks (HsImportDecl (SrcLoc _ _ row col) modName qual as (Just (_, ents)))
         = let startPos = (row, col)
               endPos = if ents == [] then if isJust as then  snd (startEndLoc toks (fromJust as))
                                                        else  snd (startEndLoc toks modName)
                                      else snd (startEndLoc toks (glast "startEndLocImport" ents))
           in extendBackwards toks startPos endPos isCloseBracket 


instance StartEndLoc  [HsExportSpecI ModuleName PNT] where
   startEndLoc toks es 
     = if es == [] then (simpPos0, simpPos0)
                   else let (startLoc, _) = startEndLoc toks $ head es 
                            (_, endLoc)   = startEndLoc toks $ last es
                        in (startLoc, endLoc)
                        -- in extendBothSides toks startLoc endLoc isOpenBracket isCloseBracket 


instance StartEndLoc (HsExportSpecI ModuleName PNT) where
     startEndLoc toks (EntE ent) =startEndLoc toks ent
    
     startEndLoc toks (ModuleE moduleName) = let (startPos, endPos) = startEndLoc toks moduleName
                                             in extendForwards toks startPos endPos isModule

    
              
instance StartEndLoc(EntSpec PNT) where
      startEndLoc toks (Var i)=startEndLoc toks i   --- x (a variable identifier)
 
      startEndLoc toks (Abs i) =startEndLoc toks i   -- T, C

      startEndLoc toks (AllSubs i) =let (startPos, endPos) =startEndLoc toks i -- T(..), C(..)
                                    in extendBackwards toks startPos endPos isCloseBracket
      startEndLoc toks (ListSubs i ents)= let (startPos, _) = startEndLoc toks i --T (C_1, ...,C_n, f1,...f_n)
                                              (_, endPos)   = startEndLoc toks (glast "startEnPosListSubs" ents)
                                          in extendBackwards toks startPos endPos isCloseBracket
                                        
instance StartEndLoc ModuleName where
   startEndLoc toks (SN modName (SrcLoc _ _ row col)) = ((row,col), (row,col))

instance StartEndLoc [EntSpec PNT] where
      startEndLoc toks ents 
        = if ents==[] then (simpPos0,simpPos0)
                      else let (startPos, _)=startEndLoc toks $ head ents
                               (_,  endPos) =startEndLoc toks $ last ents 
                           in (startPos,endPos)
                --         in extendBothSides toks startPos endPos isHiding isCloseBracket

instance StartEndLoc PNT where    
     startEndLoc toks pnt =
        case pnt of 
          PNT pn  _ (N (Just (SrcLoc _ _ row col)))->((row,col),(row,col))
          _                                        ->(simpPos0,simpPos0)  {-Shouldn't cause any problems here, as in a normal
                                                                            AST, every PNT has a source location. -}

instance StartEndLoc HsDeclP where      
    startEndLoc toks t       
       = let startPos1 = startLocOfDecl t
             (startPos2 ,endPos)=startEndLoc' t
             startPos = min  startPos1 startPos2  
             (toks1,toks2)=break (\t->tokenPos t==startPos) toks
             (toks21,toks22)=break (\t->tokenPos t>endPos) toks2 
             brackets1=filter isOpenBracket $
                        takeWhile (\t->isWhite t || isOpenBracket t)  (reverse toks1) 
             startPos'=if brackets1==[] then startPos else (tokenPos.(glast "startEndLoc:HsDeclP")) brackets1 
             brackets2=filter (not.isWhite)$
                         takeWhile (\t->tokenCol t>snd startPos && (not (isIn t)) && (not (isComma t))
                                        || isWhite t) toks22
             brackets2'=filter (not.isWhite) $ dropCloseSqBrackets (extraCloseSqBrackets (toks21++brackets2)) brackets2
             endPos'=if brackets2'==[] then endPos  else (tokenPos.(glast "startEndLoc:HsDeclP")) brackets2'
         in (startPos',endPos')
     where              
         startEndLoc' t= let locs=srcLocs t 
                         in (ghead "StartEndLoc:HsDeclP" locs, glast "StartEndLoc:HsDeclP" locs)
 
         extraCloseSqBrackets toks=length (filter isCloseSquareBracket toks)-length (filter isOpenSquareBracket toks)
         dropCloseSqBrackets 0 toks=toks
         dropCloseSqBrackets n toks=let (toks1,toks2)=break isCloseSquareBracket (reverse toks)
                                    in dropCloseSqBrackets (n-1) (reverse (tail toks2)) 
 
         startLocOfDecl :: HsDeclP -> SimpPos
         startLocOfDecl (Dec (HsTypeDecl (SrcLoc _ _ r c) _ _)) = (r,c)
         startLocOfDecl (Dec (HsDataDecl (SrcLoc _ _ r c) _ _ _ _ )) = (r,c)
         startLocOfDecl (Dec (HsNewTypeDecl (SrcLoc _ _ r c) _ _  _ _)) =(r,c)
         startLocOfDecl (Dec (HsClassDecl (SrcLoc _ _ r c) _ _ _ _))=(r,c)
         startLocOfDecl (Dec (HsInstDecl (SrcLoc _ _ r c) _ _ _ _)) = (r,c)
         startLocOfDecl (Dec (HsDefaultDecl (SrcLoc _ _ r c) _)) = (r,c)
         startLocOfDecl (Dec (HsTypeSig (SrcLoc _ _ r c) _ _ _  )) = (r, c)
         startLocOfDecl (Dec (HsFunBind (SrcLoc _ _ r c) _)) = (r,c)
         startLocOfDecl (Dec (HsPatBind (SrcLoc _ _ r c) _ _ _ )) = (r,c)
         startLocOfDecl (Dec (HsInfixDecl (SrcLoc _ _ r c) _ _)) = (r,c)

---------------End of the class StartEndLoc----------------------------------------

-- these functions have problems when they encounter sth. like [.....[p]]/
extendBothSides  toks startLoc endLoc  forwardCondFun backwardCondFun
       =let (toks1,toks2)=break (\t->tokenPos t==startLoc) toks
            toks21=dropWhile (\t->tokenPos t<=endLoc) toks2
            firstLoc=case (dropWhile (not.forwardCondFun) (reverse toks1)) of 
                             [] -> startLoc    -- is this the correct default?
                             l  -> (tokenPos.ghead "extendBothSides:lastTok") l  
            lastLoc =case (dropWhile (not.backwardCondFun) toks21) of
                            [] ->endLoc   --is this a correct default?
                            l -> (tokenPos.ghead "extendBothSides:lastTok") l                    
        in (firstLoc, lastLoc)   

extendForwards toks startLoc endLoc forwardCondFun
       =let toks1=takeWhile (\t->tokenPos t<startLoc) toks          
            firstLoc=case (dropWhile (not.forwardCondFun) (reverse toks1)) of
                       [] ->startLoc  -- is this the correct default?
                       l -> (tokenPos.ghead "extendForwards") l
        in (firstLoc, endLoc)

extendBackwards toks startLoc endLoc backwardCondFun
       = let toks1=dropWhile (\t->tokenPos t<=endLoc) toks
             lastLoc=case (dropWhile (not.backwardCondFun) toks1) of
                          [] ->endLoc -- is this the correct default?
                          l ->(tokenPos. ghead "extendBackwards") l
         in (startLoc, lastLoc)     
                                   
------------------Some functions for associating comments with syntax phrases.---------------------------
{- Note: We assume that a comment before t belongs to t only if there is at most one blank line between them, 
         and a cooment after t belongs to t only it the comment starts at the last line of t.
-}

{-Get the start&end location of syntax phrase t, then extend the end location to cover the comment/white spaces
  or new line which starts in the same line as the end location-}
startEndLocIncFowComment::(Term t, Printable t,StartEndLoc t)=>[PosToken]->t->(SimpPos,SimpPos)
startEndLocIncFowComment toks t
       =let (startLoc,endLoc)=getStartEndLoc toks t
            toks1=dropWhile (\t->tokenPos t<=endLoc) toks
            (toks11,tok12)=break (\t->tokenRow t/=fst endLoc) toks1
        in  if toks11/=[] && all (\t->isWhite t || endsWithNewLn t) toks11 
              then (startLoc, tokenPos (glast "startEndLocIncFowComment" toks11))
              else (startLoc, endLoc) 


{-get the start&end location of t in the token stream, then extend the end location to cover
  the following '\n' if there is no other characters (except white space) between t and the '\n'
-}
startEndLocIncFowNewLn::(Term t, Printable t,StartEndLoc t)=>[PosToken]->t->(SimpPos,SimpPos)
startEndLocIncFowNewLn toks t
  =let (startLoc,endLoc)=getStartEndLoc toks t
       toks1=dropWhile (\t->tokenPos t<=endLoc || isWhiteSpace t) toks
       nextTok= if toks1==[] then defaultToken else head toks1
   in if isNewLn nextTok  
        then (startLoc, tokenPos nextTok)
        else (startLoc, endLoc)
 
{-get the start&end loation of t in the token stream, then extend the start and end location to
  cover the preceding and folllowing comments. 
-}
startEndLocIncComments::(Term t, StartEndLoc t,Printable t)=>[PosToken]->t->(SimpPos,SimpPos)
startEndLocIncComments toks t
  =let (startLoc,endLoc)=getStartEndLoc toks t 
       -- tokens before 't'         
       (toks11,toks12)=break (\t->tokenRow t==fst startLoc) toks  
       toks12'=takeWhile (\t->tokenPos t <startLoc) toks12
       startLoc'=
         if all isWhite  toks12'  
           then  -- group the toks1 according to lines in a reverse order.
                 let  groupedToks=groupBy (\x y->tokenRow x==tokenRow y) (reverse toks11)
                      -- empty lines right before t
                      emptyLns=takeWhile (all (\t->isWhiteSpace t || isNewLn t )) groupedToks             
                      lastComment=if length emptyLns <=1  -- get the comment if there is any
                                    then takeWhile (all isWhite) $ takeWhile (any isComment) $ dropWhile 
                                               (all (\t->isWhiteSpace t || isNewLn t)) groupedToks    
                                    else [] -- no comment 
                      toks1'=if lastComment /=[] then reverse $ concat (emptyLns ++ lastComment)
                                                 else []
                 in if toks1'==[]
                       then if toks12'/=[] 
                              then (tokenPos (ghead "startEndLocIncComments"  toks12'))  --there is no comment before t
                              else startLoc 
                       --there is a comment before t
                       else tokenPos (ghead "startEndLocIncComments"  toks1') 
           else startLoc          
       -- tokens after t
       toks2=dropWhile (\t->tokenPos t<=endLoc) toks
       -- toks21 are those tokens that are in the same line with the last line of t 
       (toks21,tok22)=break (\t->tokenRow t/=fst endLoc) toks2
   in if toks21==[] then (startLoc',endLoc)  -- no following comments.
        else if all (\t->isWhite t || endsWithNewLn t) toks21 --get the following white tokens in the same
                                                              --line of the last token of t
               then (startLoc', tokenPos (last toks21))
               else (startLoc', endLoc)

-------------------------------------------------------------------------------------------------
astToSource=lines.replaceTabBySpaces.render.ppi
    where
       --Replace Tab by white spaces. (1 Tab=8 white spaces)
       replaceTabBySpaces::String->String
       replaceTabBySpaces []=[]
       replaceTabBySpaces (s:ss)
           =if s=='\t' then replicate 8 ' ' ++replaceTabBySpaces ss
                       else s:replaceTabBySpaces ss


 

