module PfeBrowserF(mainF) where
import Prelude hiding (putStr,putStrLn,print,readFile,ioError,catch,readIO)
import Monad(unless,mplus,when,filterM)
import Maybe(fromMaybe,mapMaybe,isJust)
import List(find,(\\),nub)
import AllFudgets as Fud hiding (getModificationTime)
import ContribFudgets hiding (lift)

import FudgetIOMonad1
import PfePlumbing
import PfeBrowserMonad
import IdInfo
import PfeBrowserGUI
import CertAttrsEditorF
import PfeBrowserMenu
import PfeBrowserCerts
import CertServerF
import AuxWindowsF
import ImageGraphics(gifFile)
import FudExtras(longTextPopupF)

import HsModule(ModuleName(..),hsModName,{-,HsModuleI-})
import HsTokens as T
import PropPosSyntax(getHSName,SrcLoc(..))
import TiDecorate(no_use)
import TI(Typing(..),envFrom)
import PNT(PNT(..))
import TypedIds(namespace,isClassOrType)
import UniqueNames as UN(OptSrcLoc(N),optOrigModule)
import HasBaseName(getBaseName)

import PFE0(getModuleInfixes,setBatchMode,checkProject,ppFlags,
	    updateModuleGraph,findNode,projectDir,getCurrentModuleGraph)
import PFE2(getModuleExports)
import PFE3(parseSourceFile'')
import PFE4(typeCheck)
import Pfe2Cmds(ppExports)
import Pfe4Cmds(ppIface,moduleInterface)
import PFE_Certs
import CertServers(printAttrs)

import PPU(pp,ppi,ppu,withPPEnv,($$),(<>),(<+>))
import DrawDoc(drawDoc)
import DrawLex(origLabel,certIcon)
import MT(lift)
import MUtils
import AbstractIO hiding (getEnv)
import FileUtils(updateFile)
import Graph(reverseGraph,neighbours)

-- Debuging
--import IOExts(trace)

--- Plumbing -------------------------------------------------------------------

mainF opts = loopOnlyF $ titleShellF' (setSizing Static) myname pfeBrowserF
  where
    pfeBrowserF = pfeF opts>==<guiF

-- @ guiF : the graphical user interface fudget of the PFE browser
guiF :: F Out In
TagF guiF dispatch tags =
    mapTF vBoxF $
        (pfeMenuBarTF >&< infoWindowsTF)
    >&< compTagF (hSplitF' aLeft) graphDispTF moduleDispTF
    >&< (popupsTF >&< certServerTF)

toTitle = Left
(_toMenuBar:&:_toInfoWindows):&:(toGraph:&:toModuleDisp)
 :&:(toPopups:&:toCertServer)=
    extend Right tags
toMessagePopup:&:toNewCertPopup:&:toCertAttrsEditors = toPopups

pfeMenuBarTF = tagF fromMenuBar pfeMenuBarF
popupsTF = messagePopupTF >&< newCertPopupTF >&< certAttrsEditorsTF
  where
    messagePopupTF = tagF fromMessagePopup (snd>^=<longTextPopupF)
    newCertPopupTF = tagF fromNewCertPopup newCertPopupF
    certAttrsEditorsTF = tagF fromCertAttrsEditor dynCertAttrsEditorPopupsF
graphDispTF = tagF fromGraph graphDispF
certServerTF = tagF fromCertServer certServerF -- actually not a GUI part...

toNodeInfo:&:(_toSource:&:_toRefs) = toModuleDisp
moduleDispTF =
    mapTF vBoxF $     tagF fromNodeInfo nodeInfoF
                  >&< compTagF (vSplitF' aBottom)
		               (tagF fromSource sourceDispF)
			       (tagF fromRefs refDispF)

infoWindowsTF = tagF fromInfoWindows $
	        spacer1F (noStretchS False True `compS` rightS) $
		auxGfxWindowsF myname

fromGraph = newNode
fromNodeInfo = either newModule newFile
fromRefs = infoPick Nothing
fromMessagePopup _ = done

fromInfoWindows = either popupInfoWindow infoPick'
  where infoPick' (w,infolbl) = infoPick (Just w) infolbl

infoPick optw infolbl =
  case infolbl of
    CertLbl cert -> clearRefs >> changeHilite [] >> showCertInfo cert
    CertCmd cmd ->
        do maybe done close optw
	   putCertServer cmd
      where close w = popupInfoWindow (w,False)
    CreateCert ca -> newCertPopup ca
    CertEdit e -> pfePut (toCertAttrsEditors (Right e))
    CertShowText txt -> pfePut . toMessagePopup . lines $ txt

newCertPopup ca =
  do cicons <- certIcons # getStBr
     clearRefs
     pfePut (toNewCertPopup (cicons,ca))

putCertServer cmd =
  do pfePut (toRefs (linesD [g "Sent certificate server command:",
			     g cmd,
			     g "Wait for result to appear"]))
     pfePut (toCertServer cmd)

fromNewCertPopup (ty,(cert,concl)) =
  if null cert || null concl
  then popupMsg "Fill in both a certificate name and a conclusion when creating a new certificate!"
  else do m <- modname # getStBr
          let cmd = unwords ["cert new",ty,quote (pp m++"/"++cert),quote concl]
          popupInfoWindow (CertInfo,False)
	  putCertServer cmd

fromCertAttrsEditor (ctype,(cname,userattrs)) =
  do Just dir <- projectDir
     m <- modname # getStBr
     (_,(Just oldattrs,_)) <- certInfo m cname
     let newattrs = updateAttrs oldattrs userattrs
     updateFile (certAttrsPath m cname dir) (printAttrs newattrs)
     infoPick Nothing (CertLbl cname) -- hmm
 where
    updateAttrs oldattrs newattrs =
        map replace oldattrs++addedattrs -- preserve the order!
      where
	addedattrs = filter added newattrs
        added (n,_) = n `notElem` map fst oldattrs
        replace (name,old) = (name,fromMaybe old (lookup name newattrs))

fromMenuBar msg = 
  case msg of
    File fcmd -> fileCmd fcmd
    View vcmd -> viewCmd vcmd
    Windows wcmd -> windowCmd wcmd
    Cert ccmd -> certCmd ccmd
    _ -> done -- !!
  where
    fileCmd Quit = pfeQuit
    fileCmd ReloadModule = reloadProject

    viewCmd vcmd = do updStBr $ \ st->st{viewMode=vcmd}
		      optGetTypes

    windowCmd (WindowCmd w up) =
        pfePut (toInfoWindows (w,Left up)) --up is False or True

    certCmd certcmd =
      case certcmd of
        NewCert -> unlessM (isNoModule.modnode # getStBr) (newCertPopup ("",""))
	TodoCert -> pfePut (toCertServer "cert todo")
	RevalidateAllCert -> revalidateAll >> reloadModule

	_ -> done

fromCertServer (ok,output) =
  do clearRefs
     changeHilite []
     {-when ok-}
     reloadModule
     unless (null output) (pfePut . route . map stripEither $ output)
  where
    route msg =
      if ok && length msg <= 2
      then toRefs . linesD . map g $ msg
      else toMessagePopup msg

--- Initialization and main loop -----------------------------------------------

--pfeF o = nullF
pfeF :: Opts -> PfeF
pfeF = runFIOM . runPfeFM pfeFM

-- pfeFM :: PfeFM ()
pfeFM =
    do g <- getCurrentModuleGraph
       pdir <- getProgramaticaDir
       servers <- getCertServers' pdir
       icons <- getCertIcons servers
       sadicon <- getIcon (gifFile (badCertIconPath pdir))
       setStBr St{modnode=noModuleNode,viewMode=view0,mrefs=[], hilbls=[],
                 types=Nothing,certs=[],revgraph=reverseGraph (map snd g),
		 certServers=servers,certIcons=icons,sadIcon=sadicon}
       pfePut (toGraph g)
       putInfoWindow (CertTypes,drawCertTypes (zip servers icons))
       pfePut (toCertAttrsEditors (Left servers))
       setBatchMode False
       pfeloop
  where	  
    pfeloop :: PfeFM ()
    pfeloop =
      do (dispatch =<< pfeGet) `catch` stdHandler
	 pfeloop

    getCertIcons :: CertServers -> PfeFM CertIcons
    getCertIcons servers = mapM getCertIcon servers
    getCertIcon certServer =
      do let dir = serverDir certServer
	     certtype = serverName certServer
             iconsrc status = gifFile (dir++"/"++status++".gif")
	 valid   <- getIcon (iconsrc "valid")
	 invalid <- getIcon (iconsrc "invalid")
	 unknown <- getIcon (iconsrc "unknown")
	 return (certtype,(valid,invalid,unknown))

    getIcon = lift . xrequestFIOM . convToPixmapK

    drawCertTypes = htableD hdrs . concatMap certD
      where
	hdrs = ["Icon","Type","Description"]
	certD (server,(ty,(icon,_,_))) = [g icon,g ty,g descr]
	  where
	    descr = fromMaybe "(Descrption missing)" $
		    lookup "description" (serverAttrs server)

typeErrorHandler err = setNoTypes >> stdHandler err

setNoTypes = do updStBr $ \ st->st{viewMode=Hyper}
                putStrLn "Switching off type info"
                pfePut (toMenuBar (View Hyper))

stdHandler err = putStrLn msg >> popupMsg msg
  where msg = ioeGetErrorString err

--- Reactions and auxiliary functions ------------------------------------------

reloadProject = reloadGraph >> reloadModule
  where
    reloadGraph =
      do updateModuleGraph
	 -- should also check for added/removed files and update graphDispF
	 updModNode . modname =<< getStBr

updModNode m =
  do g <- getCurrentModuleGraph
     node <- findNode m
     updStBr $ \st->st{revgraph=reverseGraph (map snd g),modnode=node}
     return node

reloadModule =
  do node <- modnode # getStBr
     unless (isNoModule node) (newNode node)

{-+ #newNode#: load and display a new module, after the user has entered or
     clicked on a module name or a file name. -}
newNode node@(path,_) =
  withWaitCursor $
  do dir <- checkProject
     --pfePut (toSource (Left ("","",[]))) -- empty window to scroll to top
     r <- try (parseSourceFile'' path)
     case r of
       Left err -> badNode dir err
       Right (ts,((wm,ast),refs)) ->
        do node@(_,(m,_)) <- updModNode (getBaseName (hsModName ast)) --updates revgraph
	   let --refs =simplifyRefsTypes' rs
	       asserts = assertions ast
	       certs1 = mapMaybe isCertToken ts
	       isCert cert = doesFileExist (certAttrsPath m cert dir)
	   certs2 <- filterM isCert . fromMaybe [] =<<
			maybeM (getDirectoryContents (certModuleDir m dir))
	   let certs = nub (certs1++certs2)
	   certsStatus <- zip certs # quickCertsStatus m certs dir
	   certsInfo <-
              let replace = zipWith replace1 certsStatus
                  replace1 (_,s) (c,(a,s0)) = (c,(a,s))
              in replace # mapM (certInfo m) certs
	   updStBr $ \st->st{mrefs=refs,hilbls=[],types=Nothing,certs=certsInfo}
	   e <- getModuleExports m
	   -- To keep the displays consistent with the internal state,
	   -- output operations should come after all operations that can fail!
	   clearRefs
	   importedby <- flip neighbours m . revgraph # getStBr
	   pfePut (toNodeInfo (node,importedby))
	   cicons <- icons # getStBr
	   pfePut (toSource (Left (dir,cicons,node,ts,refs,certsStatus)))
	   o <- ppFlags
	   putInfoWindow (Exports,ppo o (ppExports m e))
	   putInfoWindow (Interface,ppMsgGfx o clickToActivate)
	   putInfoWindow (Pretty,ppo o ast)
	   putInfoWindow (Evidence,evidenceGfx o dir cicons certsInfo asserts)
	   pfePut (toTitle (myname++": "++pp m))
	   optGetTypes
  where
    {-+
    Rather than aborting newNode (and potentially leaving the history buttons
    in an incosistent state) on errors, display an empty window and continue...
    -}
    badNode dir err =
      do o <- ppFlags
	 s <- fromMaybe [] # maybeM (readFile path)
         updStBr $ \st->st{modnode=node}
	 pfePut (toRefs (ppMsgGfx o (ioeGetErrorString err)))
	 pfePut (toNodeInfo (node,[]))
	 pfePut (toSource (showsrctext s))
	 putInfoWindow (Exports,blankD 10)
	 putInfoWindow (Interface,blankD 10)
	 putInfoWindow (Pretty,blankD 10)
	 putInfoWindow (Evidence,blankD 10)
	 pfePut (toTitle (myname++": "++path))


{-+ #newModule#: load and display a new module after the user has entered
    or clicked on a module name. -}
newModule :: ModuleName ->PfeFM ()
newModule m = newNode =<< findNode m

{-+ #newFile#: load and display a new module after the user has entered or 
    clicked on a file name. -}
newFile :: FilePath ->PfeFM ()
newFile path =
  do updateModuleGraph
     g <- getCurrentModuleGraph
     case [node|node@(path',_)<-g,path'==path] of
       [node] -> newNode node
       _ -> fail ("File not part of this project:"++path)

{-
goto o@(_,r) =
  case r of 
    Left (s,n) ->
      do let m' = Module s
         m <- modname # getStBr
	 unless (m'==m) $ newModule m'
	 gotoLocal o
    _ -> gotoLocal o
-}

goto o@(r,_) =
      do let optm' = optOrigModule r
         m <- modname # getStBr
	 case optm' of
           Just m'| m'/=m -> newModule m'
	   _ -> done
	 gotoLocal o

changeHilite new =
  do old <- hilbls # getStBr
     pfePut (toSource (hilite (old\\new) (new\\old)))
     updStBr $ \st->st{hilbls=new}

gotoLocal (o,sp) =
  do st <- getStBr
     case [r|r@((d,_),_,[(o',sp')])<-mrefs st,
             isDef d,o'==o,namespace sp'==namespace sp]
       of orig:_ -> do let lbl = origLabel orig
                       pfePut (toSource (mkvis lbl))
		       showInfo lbl
          _ -> putStrLn $ "Didn't find def of "++show o

getIdInfo lbl@(Lbl token) =
  do refs <- mrefs # getStBr
     infixes <- getModuleInfixes
     path <- fst . modnode # getStBr
     let p = (path,lblPos lbl)
	 idinfo = idInfo refs infixes token
     return (idinfo,p)

fromSource (Just (Button 1),lbl@(Lbl ((ModuleName,(_,mn)),_))) =
    newModule (PlainModule mn) -- !!!
fromSource (Just button,lbl@(Lbl token)) =
    case button of
      Button 1 -> case isCertTokenRef token of
	            Just cert -> changeHilite [lbl] >> showCertInfo cert
	            _ -> do info@((_,optdef,_),_) <- getIdInfo lbl
			    case optdef of
			      Nothing -> done
			      Just def ->
		                if isDefLbl lbl
			        then showInfo' lbl info
			        else goto def
      Button 2 -> showInfo lbl
      Button 3 -> do ((occurences,_,refinfo),_) <- getIdInfo lbl
		     let cnt=length occurences
			 msg=if cnt>0
			     then ["Highlighted all "++show cnt++
				    " occurences in this module"]
		             else []
                     changeHilite (map posLabel occurences)
	             pfePut (toRefs (linesD (map g (refinfo++msg))))
      _ -> done
fromSource (Nothing,lbl) =
    ifM (isLink lbl)
        (pfePut (toSource setlinkcursor))
        (pfePut (toSource setnormalcursor))
  where
    isLink (Lbl ((ModuleName,(_,mn)),_)) = return True
    isLink lbl = 
      if isDefLbl lbl
      then return False
      else do info@((_,optdef,_),_) <- getIdInfo lbl
	      return (isJust optdef)


showInfo lbl = showInfo' lbl =<< getIdInfo lbl

showInfo' lbl ((_,def,refinfo),p) =
    do optGetTypes
       tinfo <- maybeD (optGetType p) def
       ainfo <- maybeD getAInfo def
       pfePut (toRefs (linesD (map g refinfo++ainfo++tinfo)))
       changeHilite (if isJust def then [lbl] else [])
  where
    maybeD = maybe (return [])

    optGetType p o =
      ifM (haveTypes # getStBr)
          (getType p o)
          (return [])

    getType p (qn,k) =
      do --print (k,qn)
         m <- modname # getStBr
         env <- modtypes m # retypeCheck m
	 o <- ppFlags
	 --print (let xs:>:_ = unzipTyped mt in xs)
	 return [typeInfo o env k p qn]

    typeInfo ppopts (mt,env) k p qn =
	if isClassOrType k
	then kinfo qn (fst env)
	else tinfo (snd (no_use env)) mt p qn
     where
	kinfo qn = optpp . lookup qn . convTEnv (ppi.fst)

	optpp = maybe (Fud.g "No type/kind info available") (ppo ppopts)

	tinfo env mt (path,lexpos) qn =
            optpp (find' mt `mplus` lookup qn env')
	  where env' = convTEnv (ppi.fst) (env++mt)
		qn' = getHSName qn
                find' = fmap pp_use . find (pos.name)
		name (i:>:_) = getHSName i
		pos (PNT pn _ (N (Just (SrcLoc f _ lin col)))) =
                  pn==qn' && f==path && col==column lexpos && lin==line lexpos
		pos i = False
		pp_use (_:>:(gen,opti)) =
		  case opti of
		    Just inst -> "Generally:"<+>gen<>","$$"Here:"<+>inst
		    _ -> ppi gen

                pntpos (PNT _ _ (N p)) = p

    convTEnv f env = [(bn i,f info)|i:>:info<-env]
      where
        bn = fmap orig -- . getHSName
        orig (PNT n _ _) = n

    modtypes m pfe4info = (snd (envFrom tm),(concat kenvs,concat tenvs))
      where
        Just (_,(Just tm,_)) = lookup m (pfe4info::TInfo)
	(kenvs,tenvs) = unzip [env|(_,(_,(_,(_,env))))<-pfe4info]

optGetTypes =
  do st <- getStBr
     when (viewMode st==Typed && not (isNoModule (modnode st)))
	  (retypeCheck (modname st)>>done)
  `catch` typeErrorHandler

retypeCheck m =
  do opttypes <- types # getStBr
     case opttypes of
       Just types -> return types
       Nothing ->
	 withWaitCursor $
	 do types <- typeCheck (Just [m])
	    mIface <- moduleInterface m -- calls typeCheck again ==> slow!!
	    o <- ppFlags
	    putInfoWindow (Interface,ppo o (ppIface mIface))
	    updStBr $ \st->st{types=Just types}
	    return types

clearRefs = pfePut (toRefs (blankD 5))

ppMsgGfx o = linesD . map g . msglines . ppu o
--msgGfx = vboxlD' 0 . map (atomicD . expandTabs 8)

ppo ppopts = ppo' ppopts
ppo' (_,ppopts) = fontD ppfont . drawDoc . withPPEnv ppopts . ppi

msglines = map (expandTabs 8) . lines
popupMsg = pfePut . toMessagePopup . msglines

clickToActivate = 
--  "(To activate type checker: click with the middle mouse button on an identifier)"
  "(To activate type checker: select Type Info from the View menu)"

---
{-
instance CatchIO e m => CatchIO e (WithState s m) where
  stm `catch` h =  do s <- getSt
		      r <- lift $ try (withStS s stm)
		      case r of
		        Left err -> h err
			Right (ans,s') -> setSt_ s' >> return ans
  ioError = lift . ioError
-}

--
ppfont = argKey "ppfont" refFont
